use std::{
    collections::HashMap as StdHashMap,
    env,
    path::{Path, PathBuf},
    sync::Mutex,
    thread,
};

#[cfg(feature = "async")]
use smol::future;

#[cfg(feature = "async")]
use puteketeke::Executor;

#[cfg(feature = "async")]
use once_cell::sync::OnceCell;

#[cfg(feature = "tracy-client")]
use tracy_client::{non_continuous_frame, span, Client};

use abi_stable::{
    library::{lib_header_from_path, LibrarySuffix, RawLibrary},
    std_types::{RBox, RErr, ROk, ROption, RResult, Tuple2},
};
use ansi_term::Colour;
use crossbeam::channel::{unbounded, Receiver, Sender};
use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
use snafu::{location, Location};

use crate::{
    bubba::{
        error::{BubbaError, Error, Result},
        value::Value,
        RESULT, STRING,
    },
    chacha::{
        ffi_value::FfiValue,
        value::{Enum, Struct, TupleEnum},
    },
    keywords::{INVOKE_FUNC, INVOKE_FUNC_MUT},
    lu_dog::{ValueType, ValueTypeEnum},
    new_ref,
    plug_in::{Error as FfiError, LambdaCall, PluginModRef, PluginType},
    s_read, s_write,
    sarzak::{ObjectStore as SarzakStore, Ty, MODEL as SARZAK_MODEL},
    DwarfInteger, NewRef, RefType, Span, LAMBDA_FUNCS,
};

use super::instr::{Instruction, Program};

#[cfg(feature = "async")]
static mut EXECUTOR: OnceCell<Executor> = OnceCell::new();

static mut WRITE_MUTEX: OnceCell<Mutex<()>> = OnceCell::new();

#[derive(Debug)]
enum StackValue {
    Pointer(RefType<Value>),
    Value(Value),
}

impl StackValue {
    #[inline]
    fn into_pointer(self) -> RefType<Value> {
        match self {
            StackValue::Pointer(p) => p,
            StackValue::Value(v) => new_ref!(Value, v),
        }
    }

    #[inline]
    fn into_value(self) -> Value {
        match self {
            StackValue::Pointer(p) => s_read!(p).clone(),
            StackValue::Value(v) => v,
        }
    }
}

impl PartialEq for StackValue {
    fn eq(&self, other: &Self) -> bool {
        match (self, other) {
            (StackValue::Pointer(p1), StackValue::Pointer(p2)) => &*s_read!(p1) == &*s_read!(p2),
            (StackValue::Value(v1), StackValue::Value(v2)) => v1 == v2,
            _ => false,
        }
    }
}

impl Clone for StackValue {
    fn clone(&self) -> Self {
        match self {
            StackValue::Pointer(p) => StackValue::Pointer(p.clone()),
            StackValue::Value(v) => StackValue::Value(v.clone()),
        }
    }
}

impl std::fmt::Display for StackValue {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            StackValue::Pointer(p) => write!(f, "{}", *s_read!(p)),
            StackValue::Value(v) => write!(f, "{}", v),
        }
    }
}

impl From<RefType<Value>> for StackValue {
    fn from(p: RefType<Value>) -> Self {
        StackValue::Pointer(p)
    }
}

impl From<Value> for StackValue {
    fn from(v: Value) -> Self {
        StackValue::Value(v)
    }
}

#[derive(Clone)]
pub struct VM {
    instrs: Vec<Instruction>,
    source_map: Vec<Span>,
    func_map: HashMap<String, (usize, usize)>,
    sarzak: SarzakStore,
    args: RefType<Value>,
    home: PathBuf,
    captures: Option<Vec<RefType<Value>>>,
    program: Program,
    labels: HashMap<String, usize>,
    #[cfg(feature = "async")]
    thread_count: usize,
    backtrace: bool,
    lambda_sender: Sender<LambdaCall>,
    lambda_receiver: Receiver<LambdaCall>,
    trace: bool,
}

impl std::fmt::Debug for VM {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        writeln!(f, "func_map: {:?}", self.func_map)?;
        writeln!(f, "args: {:?}", self.args)?;
        writeln!(f, "home_dir: {:?}", self.home)?;
        writeln!(f, "captures: {:?}", self.captures)?;
        writeln!(f, "labels: {:?}", self.labels)?;
        Ok(())
    }
}

impl VM {
    #[cfg(feature = "async")]
    pub fn new(
        program: &Program,
        args: &[RefType<Value>],
        home: &Path,
        thread_count: usize,
        trace: bool,
    ) -> Self {
        Self::inner_new(program, args, home, thread_count, trace)
    }

    #[cfg(not(feature = "async"))]
    pub fn new(program: &Program, args: &[RefType<Value>], home: &Path, trace: bool) -> Self {
        Self::inner_new(program, args, home, 0, trace)
    }

    fn inner_new(
        program: &Program,
        args: &[RefType<Value>],
        home: &Path,
        thread_count: usize,
        trace: bool,
    ) -> Self {
        #[cfg(feature = "tracy-client")]
        Client::start();
        // println!("{}", program);
        // dbg!(&program);
        let Some(Value::ValueType(str_ty)) = program.get_symbol(STRING) else {
            panic!("No STRING symbol found.")
        };

        if trace {
            eprintln!("{program}");
        }

        let backtrace = env::var("DWARF_BACKTRACE").is_ok();
        let (lambda_sender, lambda_receiver) = unbounded();

        let mut vm = VM {
            instrs: Vec::new(),
            source_map: Vec::new(),
            func_map: HashMap::default(),
            sarzak: SarzakStore::from_bincode(SARZAK_MODEL).unwrap(),
            // These are the arguments to the program. The type is String.
            args: new_ref!(
                Value,
                Value::List {
                    ty: new_ref!(ValueType, str_ty.clone()),
                    inner: new_ref!(Vec<RefType<Value>>, args.to_vec())
                }
            ),
            home: home.to_path_buf(),
            captures: None,
            program: program.clone(),
            labels: HashMap::default(),
            #[cfg(feature = "async")]
            thread_count,
            backtrace,
            lambda_sender,
            lambda_receiver,
            trace,
        };

        let mut tmp_mem: Vec<Instruction> = Vec::new();
        let mut i = 0;
        for thonk in program.iter() {
            tmp_mem.append(&mut thonk.instructions().to_owned());
            vm.source_map.append(&mut thonk.spans().to_owned());
            vm.func_map
                .insert(thonk.name().to_owned(), (i, thonk.frame_size()));
            i += thonk.instruction_card();
        }

        let mut missing_symbols = HashSet::default();
        // This is where we patch up the function calls.
        for (addr, instr) in tmp_mem.iter().enumerate() {
            match instr {
                Instruction::Label(name) => {
                    vm.labels.insert(name.to_owned(), addr);
                    vm.instrs.push(Instruction::Label(name.to_owned()));
                }
                Instruction::CallDestination(name) => {
                    if let Some((ip, _frame_size)) = vm.func_map.get(name) {
                        vm.instrs
                            .push(Instruction::Push(Value::Integer(*ip as DwarfInteger)));
                    } else {
                        missing_symbols.insert(name.to_owned());
                    }
                }
                Instruction::LocalCardinality(name) => {
                    if let Some((_ip, frame_size)) = vm.func_map.get(name) {
                        vm.instrs.push(Instruction::Push(Value::Integer(
                            *frame_size as DwarfInteger,
                        )));
                    } else {
                        missing_symbols.insert(name.to_owned());
                    }
                }
                _ => vm.instrs.push(instr.clone()),
            }
        }

        if !missing_symbols.is_empty() {
            panic!("Missing symbols: {:?}", missing_symbols);
        }

        // 🚧 This is such an ugly hack. There should probably be a thread pool.
        // OTOH, if this is sufficient...
        let mut vm_clone = vm.clone();
        thread::spawn(move || loop {
            vm_clone.lambda_listen();
        });

        let mut vm_clone = vm.clone();
        thread::spawn(move || loop {
            vm_clone.lambda_listen();
        });

        vm
    }

    fn lambda_listen(&mut self) {
        if let Ok(lambda_call) = self.lambda_receiver.recv() {
            let λ = match LAMBDA_FUNCS.get() {
                Some(λ) => λ,
                None => {
                    panic!("Lambda functions have not been initialized.");
                }
            };

            let λ = {
                let λ = λ.lock().unwrap();
                λ.get(&lambda_call.lambda).unwrap().clone()
            };

            // This will also have been set in the constructor. Calling this before
            // construction of a VM will panic, and that's not a terrible default.
            // 🚧 I don't love that there is only one of these for executing lambdas.
            // I guess it wouldn't be hard to make it a Vec of VMs. Let it grow, and
            // shrink as needed. Whatever as needed means.
            // let mut vm = ΛVM.get().unwrap().lock().unwrap();
            let args = lambda_call
                .args
                .iter()
                .map(|v| <FfiValue as Into<Value>>::into(v.clone()))
                .collect::<Vec<_>>();

            let result = self.invoke_lambda(&λ, &args);

            let result = match result {
                Ok(value) => ROk(<Value as Into<FfiValue>>::into(s_read!(value).clone())),
                Err(e) => RErr(FfiError::Plugin(e.to_string().into())),
            };
            lambda_call.result.send(result).unwrap();
        }
    }

    fn get_span(&self, ip: isize) -> Span {
        self.source_map[(ip - 1) as usize].to_owned()
    }

    pub fn invoke_lambda(&mut self, lambda: &Value, args: &[Value]) -> Result<RefType<Value>> {
        let mut stack = Vec::<StackValue>::new();
        // This is all pretty hacky. I need four 0 values to appease the return gods.
        stack.push(Value::Integer(0).into());
        stack.push(Value::Integer(0).into());
        stack.push(Value::Integer(0).into());
        stack.push(Value::Integer(0).into());
        // This is really lame. The call code expects stack_len - 2 - arity to be
        // an address (an integer) ar it doesn't do anything with it. In any case
        // it needs to be here. So this is junk.
        stack.push(Value::Empty.into());
        // The next parameter is the number of locals on the stack -or- a LambdaPointer.
        stack.push(lambda.to_owned().into());

        // Args need to be pushed onto the stack, and then the lambda, and then we
        // need to execute a call instruction.
        for arg in args.iter() {
            stack.push(arg.clone().into());
        }

        let Value::LambdaPointer {
            name,
            frame_size,
            captures: _,
        } = lambda.clone()
        else {
            panic!("Expected a lambda pointer.")
        };

        let func_name = format!("{name}_trampoline");
        let (ip, _) = self.func_map.get(&func_name).unwrap();
        self.inner_run(
            &func_name,
            *ip as isize,
            4,
            stack,
            args.len(),
            frame_size,
            self.program.clone(),
        )
    }

    pub fn invoke(&mut self, func_name: &str, args: &[RefType<Value>]) -> Result<RefType<Value>> {
        let (ip, frame_size) = self.func_map.get(func_name).unwrap();
        let frame_size = *frame_size;

        let mut stack = Vec::new();

        // Address of the function to invoke.
        stack.push(Value::Integer(*ip as DwarfInteger).into());
        // Number of parameters and locals in the function.
        stack.push(Value::Integer(frame_size as DwarfInteger).into());

        for arg in args.iter() {
            stack.push(arg.clone().into());
        }
        for _ in 0..frame_size - args.len() {
            stack.push(Value::Empty.into());
        }

        // Arity
        stack.push(StackValue::Value(
            Value::Integer(args.len() as DwarfInteger),
        ));
        // Frame size
        stack.push(StackValue::Value(Value::Integer(
            (frame_size + 2) as DwarfInteger,
        )));
        // This is the IP sentinel value.
        stack.push(Value::Empty.into());
        // Setup the frame pointer and it's sentinel.
        stack.push(Value::Empty.into());

        let fp = frame_size + 5;
        let ip = *ip as isize;

        let result = self.inner_run(
            func_name,
            ip,
            fp,
            stack,
            args.len(),
            frame_size,
            self.program.clone(),
        );

        result
    }

    fn inner_run(
        &mut self,
        // This is really just for debugging
        name: &str,
        // This is an isize because we have negative jump offsets.
        mut ip: isize,
        mut fp: usize,
        mut stack: Vec<StackValue>,
        mut arity: usize,
        mut local_count: usize,
        program: Program,
    ) -> Result<RefType<Value>> {
        #[cfg(feature = "tracy-client")]
        let _frame = non_continuous_frame!("inner_run");

        let mut cx = 0;

        loop {
            if ip as usize >= self.instrs.len() {
                return Err(BubbaError::IPOutOfBounds { ip: ip as usize }.into());
            }

            if self.trace {
                // let mutex = WRITE_MUTEX.get_or_init(|| Mutex::new(()));
                // let guard = mutex.lock().unwrap();
                let mutex = match unsafe { WRITE_MUTEX.get() } {
                    Some(mutex) => mutex,
                    None => {
                        let mutex = Mutex::new(());
                        unsafe {
                            WRITE_MUTEX.set(mutex).unwrap();
                            WRITE_MUTEX.get().unwrap()
                        }
                    }
                };
                let _guard = mutex.lock().unwrap();
                print_stack(&stack, fp);
                println!("\t{} ->\t{cx}", Colour::Green.bold().paint("cx"));
                println!("{}: {name}", Colour::Green.bold().paint("Thread"));
                print_instrs(ip, &program, &self.instrs, &self.source_map);
                println!();
            }

            let instr = &self.instrs[ip as usize];
            let ip_offset: isize = {
                match instr {
                    Instruction::Add => {
                        let b = stack.pop().unwrap();
                        let a = stack.pop().unwrap();

                        let c = a.into_value() + b.into_value();
                        stack.push(c.into());

                        1
                    }
                    Instruction::And => {
                        let b = stack.pop().unwrap();
                        let a = stack.pop().unwrap();

                        let c = Value::Boolean(
                            a.into_value().try_into()? && b.into_value().try_into()?,
                        );
                        stack.push(c.into());

                        1
                    }
                    #[cfg(feature = "async")]
                    Instruction::AsyncCall(func_arity) => {
                        #[cfg(feature = "tracy-client")]
                        let _span = span!("AsyncCall");

                        self.captures = None;

                        self.start_task(&mut stack, *func_arity, arity, &program)?;

                        1
                    }
                    #[cfg(feature = "async")]
                    Instruction::AsyncSpawn(func_arity) => {
                        #[cfg(feature = "tracy-client")]
                        let _span = span!("AsyncSpawn");

                        self.start_task(&mut stack, *func_arity, arity, &program)?;

                        1
                    }
                    #[cfg(feature = "async")]
                    Instruction::Await => {
                        #[cfg(feature = "tracy-client")]
                        let _span = span!("Await");

                        let future = stack.pop().unwrap().into_pointer();
                        let mut expression = &mut *s_write!(future);
                        // dbg!(&expression);

                        let executor = match unsafe { EXECUTOR.get() } {
                            Some(executor) => executor,
                            None => {
                                let executor = Executor::new(self.thread_count);
                                unsafe {
                                    EXECUTOR.set(executor).unwrap();
                                    EXECUTOR.get().unwrap()
                                }
                            }
                        };

                        let result = match &mut expression {
                            Value::Task {
                                name,
                                task,
                                running,
                            } => {
                                if let Some(task) = s_write!(task).take() {
                                    if !*running {
                                        tracing::trace!(target: "vm", "Starting task: {name}");
                                        executor.start_task(&task);
                                    }

                                    tracing::trace!(target: "vm", "Awaiting task: {name}");
                                    let f = future::block_on(task)?;
                                    f
                                } else {
                                    panic!(
                                        "Task ({name}) is missing -- already awaited. Why does this happen?"
                                    );
                                }
                            }
                            something_else => {
                                panic!("Expected a task, found {something_else:?}.");
                            }
                        };

                        stack.push(result.into());

                        1
                    }
                    Instruction::Call(func_arity) => {
                        self.captures = None;

                        let callee = &stack[stack.len() - func_arity - 2];
                        let stack_local_count = &stack[stack.len() - func_arity - 1];

                        if let Value::Plugin((_, plugin)) = callee.clone().into_value() {
                            let method: String =
                                stack_local_count.clone().into_value().try_into()?;

                            match method.as_str() {
                                INVOKE_FUNC => {
                                    let plugin = s_read!(plugin);
                                    let args = stack.pop().clone().unwrap().into_value();
                                    let Value::List { inner, .. } = args else {
                                        panic!("Expected a vector of arguments.")
                                    };
                                    let args = s_read!(inner)
                                        .iter()
                                        .map(|v| {
                                            <Value as Into<FfiValue>>::into(s_read!(v).clone())
                                        })
                                        .collect::<Vec<FfiValue>>();
                                    let func = stack.pop().clone().unwrap().into_value();
                                    let func = func.to_inner_string();
                                    let ty = stack.pop().clone().unwrap().into_value();
                                    let ty = ty.to_inner_string();
                                    let module = stack.pop().clone().unwrap().into_value();
                                    let module = module.to_inner_string();

                                    match plugin.invoke_func(
                                        module.as_str().into(),
                                        ty.as_str().into(),
                                        func.as_str().into(),
                                        args.into(),
                                    ) {
                                        ROk(value) => {
                                            let result = program.get_symbol(RESULT).expect(
                                                "The RESULT symbol is missing from the program.",
                                            );
                                            stack.push(
                                                <(FfiValue, &Value) as Into<Value>>::into((
                                                    value, result,
                                                ))
                                                .into(),
                                            );
                                        }
                                        RErr(e) => {
                                            return Err(BubbaError::VmPanic {
                                                message: format!("Plugin error: {:?}\nAttempting to call {module}::{ty}::{func}", e),
                                                location: location!(),
                                            }
                                            .into())
                                        }
                                    }
                                }
                                INVOKE_FUNC_MUT => {
                                    let mut plugin = s_write!(plugin);
                                    let args = stack.pop().clone().unwrap().into_value();
                                    let Value::List { inner, .. } = args else {
                                        panic!("Expected a vector of arguments.")
                                    };
                                    let args = s_read!(inner)
                                        .iter()
                                        .map(|v| {
                                            <Value as Into<FfiValue>>::into(s_read!(v).clone())
                                        })
                                        .collect::<Vec<FfiValue>>();
                                    let func = stack.pop().clone().unwrap().into_value();
                                    let func = func.to_inner_string();
                                    let ty = stack.pop().clone().unwrap().into_value();
                                    let ty = ty.to_inner_string();
                                    let module = stack.pop().clone().unwrap().into_value();
                                    let module = module.to_inner_string();

                                    match plugin.invoke_func_mut(
                                        module.as_str().into(),
                                        ty.as_str().into(),
                                        func.as_str().into(),
                                        args.into(),
                                    ) {
                                        ROk(value) => {
                                            let result = program.get_symbol(RESULT).expect(
                                                "The RESULT symbol is missing from the program.",
                                            );
                                            stack.push(
                                                <(FfiValue, &Value) as Into<Value>>::into((
                                                    value, result,
                                                ))
                                                .into(),
                                            );
                                        }
                                        RErr(e) => {
                                            return Err(BubbaError::VmPanic {
                                                message: format!("Plugin error: {:?}\nAttempting to call {module}::{ty}::{func}", e),
                                                location: location!(),
                                            }
                                            .into())
                                        }
                                    }
                                }
                                _ => {
                                    return Err(BubbaError::VmPanic {
                                        message: format!("Unknown method: {method}.",),
                                        location: location!(),
                                    }
                                    .into())
                                }
                            }

                            1
                        } else {
                            let (callee, frame_size): (isize, Value) = match stack_local_count
                                .clone()
                                .into_value()
                            {
                                Value::Integer(arity) => {
                                    let callee: isize = callee.clone().into_value().try_into()?;

                                    local_count = arity as usize;
                                    (
                                        callee,
                                        <usize as Into<Value>>::into(func_arity + local_count + 2),
                                    )
                                }
                                Value::LambdaPointer {
                                    name,
                                    frame_size,
                                    captures,
                                } => {
                                    self.captures = Some(captures);
                                    let addr = self.func_map.get(&name).unwrap().0;
                                    local_count = frame_size;
                                    (
                                        addr as isize,
                                        // This is plus one because the call frame does not include
                                        // the frame size.
                                        <usize as Into<Value>>::into(func_arity + local_count + 1),
                                    )
                                }
                                // Value::Plugin(plugin) => {}
                                _ => {
                                    return Err(BubbaError::VmPanic {
                                        message: format!(
                                            "Unexpected value: {stack_local_count:?}.",
                                        ),
                                        location: location!(),
                                    }
                                    .into())
                                }
                            };

                            let old_fp = fp;
                            let old_ip = ip;

                            // The call stack has been setup, but we need to make room
                            // for locals.
                            for _ in 0..local_count {
                                stack.push(Value::Empty.into());
                            }

                            // Push the arity
                            stack.push(<usize as Into<Value>>::into(arity).into());
                            arity = *func_arity;

                            // Push the call frame size so that we can clean in out quickly
                            stack.push(frame_size.into());

                            // Push the old IP
                            stack.push(<isize as Into<Value>>::into(old_ip).into());

                            fp = stack.len();
                            stack.push(<usize as Into<Value>>::into(old_fp).into());

                            // ip = callee;
                            // let result = self.inner_run(arity, local_count, trace)?;

                            // Move the frame pointer back
                            // fp = (&*s_read!(stack[fp])).try_into().unwrap();
                            // fp = old_fp;
                            // ip = old_ip;

                            // (0..arity + local_count + 3).for_each(|_| {
                            //     stack.pop();
                            // });

                            // stack.push(result);
                            callee - ip
                        }
                    }
                    Instruction::CaptureLocal(from, to) => {
                        let Some(captures) = self.captures.as_ref() else {
                            panic!("Attempt to capture a local outside of a lambda call.")
                        };
                        // The first one is faster, but it doesn't share captured
                        // variables across tasks.
                        // The second one does the latter, and is slower due to the
                        // locking involved.
                        // Maybe this could be configurable? Feature flag? Maybe
                        // even something at runtime, although we'd need to see
                        // how much that extra condition costs.
                        // let value = s_read!(captures[*from]).clone();
                        let value = captures[*from].clone();
                        stack[fp - arity - local_count - 3 + to] = value.into();

                        1
                    }
                    Instruction::Comment(_) => 1,
                    // Instruction::DeconstructStructExpression => {
                    //     fn decode_expression(
                    //         value: RefType<Value>,
                    //     ) -> Result<(RefType<Value>, Option<RefType<Value>>)>
                    //     {
                    //         let read = s_read!(value);
                    //         match &*read {
                    //             Value::Enumeration(value) => match value {
                    //                 // 🚧 I can't tell if this is gross, or a sweet hack.
                    //                 // I think I'm referring to using the name as the scrutinee?
                    //                 EnumVariant::Unit(_, ty, value) => Ok((
                    //                     new_ref!(Value, Value::String(ty.to_owned())),
                    //                     Some(new_ref!(Value, Value::String(value.to_owned()))),
                    //                 )),
                    //                 // EnumFieldVariant::Struct(value) => (
                    //                 //     *value.type_name().to_owned(),
                    //                 //     Some(*value.get_value()),
                    //                 // ),
                    //                 EnumVariant::Tuple((ty, path), value) => {
                    //                     let path = path.split(PATH_SEP).collect::<Vec<&str>>();
                    //                     let mut path = VecDeque::from(path);
                    //                     let name = path.pop_front().unwrap().to_owned();
                    //                     if name.is_empty() {
                    //                         Ok((
                    //                             new_ref!(
                    //                                 Value,
                    //                                 Value::String(
                    //                                     s_read!(value).variant().to_owned()
                    //                                 )
                    //                             ),
                    //                             Some(s_read!(value).value().clone()),
                    //                         ))
                    //                     } else {
                    //                         Ok((
                    //                             new_ref!(Value, Value::String(name)),
                    //                             Some(new_ref!(
                    //                                 Value,
                    //                                 Value::Enumeration(EnumVariant::Tuple(
                    //                                     (
                    //                                         ty.clone(),
                    //                                         path.into_iter()
                    //                                             .collect::<Vec<&str>>()
                    //                                             .join(PATH_SEP)
                    //                                     ),
                    //                                     value.clone(),
                    //                                 ))
                    //                             )),
                    //                         ))
                    //                     }
                    //                 }
                    //                 _ => unimplemented!(),
                    //             },
                    //             _ => Ok((value.clone(), None)),
                    //         }
                    //     }

                    //     let mut variant = stack.pop().unwrap();
                    //     while let Ok((name, value)) = decode_expression(variant.into_pointer()) {
                    //         dbg!(&name, &value);
                    //         stack.push(name.into());
                    //         if let Some(value) = value {
                    //             variant = value.into();
                    //         } else {
                    //             break;
                    //         }
                    //     }

                    //     1
                    // }
                    Instruction::Divide => {
                        let b = stack.pop().unwrap();
                        let a = stack.pop().unwrap();

                        let c = a.into_value() / b.into_value();

                        stack.push(c.into());

                        1
                    }
                    Instruction::Dup => {
                        // let value = stack.pop().unwrap();
                        // stack.push(value.clone());
                        let value = stack[stack.len() - 1].clone();
                        stack.push(value);

                        1
                    }
                    Instruction::ExtractEnumValue => {
                        let user_enum = stack.pop().unwrap();
                        let Value::Enumeration(user_enum) = user_enum.clone().into_value() else {
                            if self.backtrace {
                                eprintln!("{self:?}");
                                eprintln!("{program}");
                                print_stack(&stack, fp);
                                print_instrs(ip, &program, &self.instrs, &self.source_map);
                            }
                            return Err(BubbaError::VmPanic {
                                message: format!("Expected enum, found: {user_enum:?}."),
                                location: location!(),
                            }
                            .into());
                        };

                        match user_enum {
                            Enum::Unit(_, _, value) => {
                                stack.push(Value::String(value.to_owned()).into());
                            }
                            Enum::Tuple(_, value) => {
                                stack.push(s_read!(value).value().clone().into());
                            }
                            _ => unimplemented!(),
                        }

                        1
                    }
                    // The fp is pointing someplace near the end of the vec.
                    // Nominally at one past the Thonk name at the bottom of the stack.
                    // Any locals will cause the fp to be moved up, with the
                    // locals existing between the Thonk name and the fp.
                    Instruction::FetchLocal(index) => {
                        // We gotta index the stack from the top of the call frame.
                        let value = stack[fp - arity - local_count - 3 + index].clone();

                        // When fetching a local, we need to ensure that it's a
                        // pointer so that modifications are reflected back in the
                        // stack.
                        match value {
                            StackValue::Value(value) => {
                                let value = StackValue::Pointer(new_ref!(Value, value));
                                stack[fp - arity - local_count - 3 + index] = value.clone();

                                stack.push(value);
                            }
                            StackValue::Pointer(value) => {
                                stack.push(value.into());
                            }
                        }

                        1
                    }
                    Instruction::FieldRead => {
                        let field = stack.pop().unwrap();
                        let ty_ = stack.pop().unwrap();
                        match ty_.into_value() {
                            Value::Struct(ty_) => {
                                match ty_
                                    .get_field_value(field.clone().into_value().to_inner_string())
                                {
                                    Some(value) => {
                                        stack.push(value.clone().into());
                                    }
                                    None => {
                                        return Err::<RefType<Value>, Error>(
                                            BubbaError::NoSuchField {
                                                field: field.to_string(),
                                                ty: ty_.to_string(),
                                            }
                                            .into(),
                                        );
                                    }
                                }
                            }
                            value => {
                                if self.backtrace {
                                    eprintln!("{self:?}");
                                    print_stack(&stack, fp);
                                    print_instrs(ip, &program, &self.instrs, &self.source_map);
                                }
                                return Err::<RefType<Value>, Error>(
                                    BubbaError::VmPanic {
                                        message: format!("FieldRead unexpected value: {value}."),
                                        location: location!(),
                                    }
                                    .into(),
                                );
                            }
                        }

                        1
                    }
                    Instruction::FieldWrite => {
                        let field = stack.pop().unwrap();
                        let strukt = stack.pop().unwrap();
                        let value = stack.pop().unwrap();

                        let strukt = strukt.into_pointer();
                        let mut strukt = s_write!(strukt);
                        match &mut *strukt {
                            Value::Struct(ty) => {
                                match ty.set_field_value(
                                    field.clone().into_value().to_inner_string(),
                                    value.into_value(),
                                ) {
                                    Some(_) => {}
                                    None => {
                                        return Err::<RefType<Value>, Error>(
                                            BubbaError::NoSuchField {
                                                field: field.to_string(),
                                                ty: ty.to_string(),
                                            }
                                            .into(),
                                        )
                                    }
                                }
                            }
                            value => {
                                return Err::<RefType<Value>, Error>(
                                    BubbaError::VmPanic {
                                        message: format!("Unexpected value. type: {value}."),
                                        location: location!(),
                                    }
                                    .into(),
                                )
                            }
                        }

                        1
                    }
                    Instruction::Goto(label) => {
                        if let Some(new_ip) = self.labels.get(label) {
                            *new_ip as isize - ip
                        } else {
                            return Err(BubbaError::VmPanic {
                                location: location!(),
                                message: format!("Unknown label: {label}."),
                            }
                            .into());
                        }
                    }
                    Instruction::HaltAndCatchFire => {
                        let span = stack.pop().unwrap();
                        let span: std::ops::Range<usize> = span.into_value().try_into()?;

                        let file = stack.pop().unwrap();
                        let file: String = file.into_value().try_into()?;

                        return Err(BubbaError::HaltAndCatchFire {
                            file,
                            span,
                            ip: ip as usize,
                        }
                        .into());
                    }
                    Instruction::Incr => {
                        cx += 1;

                        1
                    }
                    Instruction::InitializeLocal(index) => {
                        let value = stack.pop().unwrap();

                        // We gotta index into the stack in reverse order from the index.
                        stack[fp - arity - local_count - 3 + index] = value;

                        1
                    }
                    Instruction::Jump(offset) => offset + 1,
                    Instruction::JumpIfFalse(offset) => {
                        let condition = stack.pop().unwrap();
                        let condition: bool = condition.into_value().try_into()?;

                        if !condition {
                            offset + 1
                        } else {
                            1
                        }
                    }
                    Instruction::JumpIfTrue(offset) => {
                        let condition = stack.pop().unwrap();
                        let condition: bool = condition.into_value().try_into()?;
                        if condition {
                            offset + 1
                        } else {
                            1
                        }
                    }
                    Instruction::Label(_) => 1,
                    Instruction::ListIndex => {
                        let index = stack.pop().unwrap().into_value();
                        let list = stack.pop().unwrap();
                        let list = list.into_pointer();
                        let list = s_read!(list);
                        let index: usize = index.try_into()?;
                        match &*list {
                            Value::AnyList(vec) => {
                                let vec = s_read!(vec);
                                if index < vec.len() {
                                    stack.push(vec[index].clone().into());
                                } else {
                                    if self.backtrace {
                                        eprintln!("{self:?}");
                                        print_stack(&stack, fp);
                                        print_instrs(ip, &program, &self.instrs, &self.source_map);
                                    }
                                    return Err(BubbaError::IndexOutOfBounds {
                                        index,
                                        len: vec.len(),
                                        span: self.get_span(ip),
                                        location: location!(),
                                    }
                                    .into());
                                }
                            }
                            Value::List { ty: _, inner: vec } => {
                                let vec = s_read!(vec);
                                if index < vec.len() {
                                    stack.push(vec[index].clone().into());
                                } else {
                                    if self.backtrace {
                                        eprintln!("{self:?}");
                                        print_stack(&stack, fp);
                                        print_instrs(ip, &program, &self.instrs, &self.source_map);
                                    }
                                    return Err(BubbaError::IndexOutOfBounds {
                                        index,
                                        len: vec.len(),
                                        span: self.get_span(ip),
                                        location: location!(),
                                    }
                                    .into());
                                }
                            }
                            Value::String(str) => {
                                let str = unicode_segmentation::UnicodeSegmentation::graphemes(
                                    str.as_str(),
                                    true,
                                )
                                .collect::<Vec<&str>>();

                                if index < str.len() {
                                    stack.push(
                                        Value::Char(
                                            str[index..index + 1].join("").chars().next().unwrap(),
                                        )
                                        .into(),
                                    )
                                } else {
                                    if self.backtrace {
                                        eprintln!("{self:?}");
                                        print_stack(&stack, fp);
                                        print_instrs(ip, &program, &self.instrs, &self.source_map);
                                    }
                                    return Err(BubbaError::IndexOutOfBounds {
                                        index,
                                        len: str.len(),
                                        span: self.get_span(ip),
                                        location: location!(),
                                    }
                                    .into());
                                }
                            }
                            value => {
                                return Err(BubbaError::NotIndexable {
                                    span: self.get_span(ip),
                                    value: value.to_owned(),
                                    location: location!(),
                                }
                                .into());
                            }
                        }

                        1
                    }
                    Instruction::ListIndexRange => {
                        let start: usize = stack.pop().unwrap().into_value().try_into()?;
                        let end: usize = stack.pop().unwrap().into_value().try_into()?;
                        let list = stack.pop().unwrap();
                        let list = list.into_pointer();
                        let list = s_read!(list);

                        match &*list {
                            Value::List { ty, inner: vec } => {
                                let vec = s_read!(vec);
                                if end < vec.len() {
                                    let list = new_ref!(
                                        Value,
                                        Value::List {
                                            ty: ty.clone(),
                                            inner: new_ref!(
                                                Vec<RefType<Value>>,
                                                vec[start..end].to_owned()
                                            )
                                        }
                                    );

                                    stack.push(list.into())
                                } else {
                                    if self.backtrace {
                                        eprintln!("{self:?}");
                                        print_stack(&stack, fp);
                                        print_instrs(ip, &program, &self.instrs, &self.source_map);
                                    }
                                    return Err(BubbaError::IndexOutOfBounds {
                                        index: end,
                                        len: vec.len(),
                                        span: self.get_span(ip),
                                        location: location!(),
                                    }
                                    .into());
                                }
                            }
                            Value::String(str) => {
                                let str = unicode_segmentation::UnicodeSegmentation::graphemes(
                                    str.as_str(),
                                    true,
                                )
                                .collect::<Vec<&str>>();

                                if end < str.len() {
                                    stack.push(Value::String(str[start..end + 1].join("")).into())
                                } else {
                                    if self.backtrace {
                                        eprintln!("{self:?}");
                                        print_stack(&stack, fp);
                                        print_instrs(ip, &program, &self.instrs, &self.source_map);
                                    }
                                    return Err(BubbaError::IndexOutOfBounds {
                                        index: end,
                                        len: str.len(),
                                        span: self.get_span(ip),
                                        location: location!(),
                                    }
                                    .into());
                                }
                            }
                            value => {
                                return Err(BubbaError::NotIndexable {
                                    span: self.get_span(ip),
                                    value: value.to_owned(),
                                    location: location!(),
                                }
                                .into());
                            }
                        }

                        1
                    }
                    Instruction::ListLength => {
                        let list = stack.pop().unwrap();
                        let list = list.into_pointer();
                        let list = s_read!(list);
                        match &*list {
                            Value::AnyList(vec) => {
                                let vec = s_read!(vec);
                                stack.push(Value::Integer(vec.len() as DwarfInteger).into());
                            }
                            Value::List { inner, .. } => {
                                let inner = s_read!(inner);
                                stack.push(Value::Integer(inner.len() as DwarfInteger).into());
                            }
                            Value::Map { inner } => {
                                let inner = s_read!(inner);
                                stack.push(Value::Integer(inner.len() as DwarfInteger).into());
                            }
                            Value::String(str) => {
                                stack.push(Value::Integer(str.len() as DwarfInteger).into());
                            }
                            value => {
                                if self.backtrace {
                                    eprintln!("{self:?}");
                                    print_stack(&stack, fp);
                                    print_instrs(ip, &program, &self.instrs, &self.source_map);
                                }
                                return Err(BubbaError::NotIndexable {
                                    span: self.get_span(ip),
                                    value: value.to_owned(),
                                    location: location!(),
                                }
                                .into());
                            }
                        }

                        1
                    }
                    Instruction::ListPush => {
                        let element = stack.pop().unwrap();
                        let list = stack.pop().unwrap();
                        let list = list.into_pointer();
                        match &*s_read!(list) {
                            Value::List { inner, .. } => {
                                let mut inner = s_write!(inner);
                                inner.push(element.into_pointer());
                            }
                            value => {
                                if self.backtrace {
                                    eprintln!("{self:?}");
                                    print_stack(&stack, fp);
                                    print_instrs(ip, &program, &self.instrs, &self.source_map);
                                }
                                return Err(BubbaError::NotIndexable {
                                    span: self.get_span(ip),
                                    value: value.to_owned(),
                                    location: location!(),
                                }
                                .into());
                            }
                        }

                        1
                    }
                    Instruction::MakeLambdaPointer(name, frame_size) => {
                        let captures: Vec<RefType<Value>> = stack
                            [fp - arity - local_count - 3..fp - 3]
                            .iter()
                            .cloned()
                            .map(|v| v.into_pointer())
                            .collect();

                        // Replace stack values with references to the captured values.
                        captures.iter().enumerate().for_each(|(i, v)| {
                            stack[fp - arity - local_count - 3 + i] = v.clone().into();
                        });

                        let name = name.to_owned();
                        let value = Value::LambdaPointer {
                            name,
                            frame_size: *frame_size,
                            captures,
                        };
                        stack.push(value.into());

                        1
                    }
                    Instruction::MapGet => {
                        let key = stack.pop().unwrap();
                        let key = key.into_value();
                        let key: String = key.try_into()?;

                        let map = stack.pop().unwrap();
                        let map = map.into_pointer();
                        let map = s_read!(map);
                        match &*map {
                            Value::Map { inner, .. } => {
                                let inner = s_read!(inner);
                                let value = inner.get(&key);
                                // stack.push(value.into());
                                match inner.get(&key) {
                                    Some(value) => {
                                        stack.push(value.clone().into());
                                    }
                                    None => {
                                        stack.push(Value::Empty.into());
                                    }
                                }
                            }
                            value => {
                                if self.backtrace {
                                    eprintln!("{self:?}");
                                    print_stack(&stack, fp);
                                    print_instrs(ip, &program, &self.instrs, &self.source_map);
                                }
                                return Err(BubbaError::NotIndexable {
                                    span: self.get_span(ip),
                                    value: value.to_owned(),
                                    location: location!(),
                                }
                                .into());
                            }
                        }

                        1
                    }
                    Instruction::MapInsert => {
                        let value = stack.pop().unwrap();
                        let value = value.into_pointer();

                        let key = stack.pop().unwrap();
                        let key = key.into_value();
                        let key = key.try_into()?;

                        let map = stack.pop().unwrap();
                        let map_value = map.into_pointer();
                        let map = s_write!(map_value);
                        match &*map {
                            Value::Map { inner, .. } => {
                                let mut inner = s_write!(inner);
                                inner.insert(key, value);
                            }
                            value => {
                                if self.backtrace {
                                    eprintln!("{self:?}");
                                    print_stack(&stack, fp);
                                    print_instrs(ip, &program, &self.instrs, &self.source_map);
                                }
                                return Err(BubbaError::NotIndexable {
                                    span: self.get_span(ip),
                                    value: value.to_owned(),
                                    location: location!(),
                                }
                                .into());
                            }
                        }

                        stack.push(map_value.clone().into());

                        1
                    }
                    Instruction::MapLength => {
                        let map = stack.pop().unwrap();
                        let map = map.into_pointer();
                        let map = s_read!(map);
                        match &*map {
                            Value::Map { inner, .. } => {
                                let inner = s_read!(inner);
                                stack.push(Value::Integer(inner.len() as DwarfInteger).into());
                            }
                            value => {
                                if self.backtrace {
                                    eprintln!("{self:?}");
                                    print_stack(&stack, fp);
                                    print_instrs(ip, &program, &self.instrs, &self.source_map);
                                }
                                return Err(BubbaError::NotIndexable {
                                    span: self.get_span(ip),
                                    value: value.to_owned(),
                                    location: location!(),
                                }
                                .into());
                            }
                        }

                        1
                    }
                    Instruction::MethodLookup(name) => {
                        let ty = stack.pop().unwrap();

                        if let Value::Plugin(_) = ty.clone().into_value() {
                            stack.push(ty);
                            stack.push(<String as Into<Value>>::into((name).clone()).into());
                        } else {
                            // We are just building the "prefix" of the function for below.
                            let ty = match ty.into_value() {
                                Value::Enumeration(variant) => match variant {
                                    Enum::Struct(ty) => {
                                        let ty = s_read!(ty);
                                        let name = ty.type_name();
                                        name.to_owned()
                                    }
                                    Enum::Tuple((_, ty), _) => ty.to_owned(),
                                    Enum::Unit(_, ty, _) => ty.to_owned(),
                                },
                                Value::Integer(_) => "::std::integer::Integer".to_owned(),
                                Value::Map { .. } => "::std::collections::HashMap".to_owned(),
                                Value::Struct(ty) => {
                                    let name = ty.type_name();
                                    name.to_owned()
                                }
                                Value::String(_) => "::std::string::String".to_owned(),
                                // Value::Vector { ty, .. } => {
                                //     let ty = s_read!(ty);
                                //     let name = ty.type_name();
                                //     name.to_owned()
                                // }
                                oopsie => panic!("{oopsie:?}"),
                            };

                            let func = format!("{}::{}", ty, name);

                            // This is probably slowing us down a lot. We need to do this
                            // lookup some other way.
                            // Couldn't this be resolved when we are compiling? I don't suppose
                            // so since it's here. Of course there's always a more clever
                            // solution.
                            if let Some((ip, frame_size)) = self.func_map.get(&func) {
                                stack.push(Value::Integer(*ip as DwarfInteger).into());
                                stack.push(Value::Integer(*frame_size as DwarfInteger).into());
                            } else {
                                return Err(BubbaError::VmPanic {
                                    message: format!("Missing function definition: {func}"),
                                    location: location!(),
                                }
                                .into());
                            }
                        }

                        1
                    }
                    Instruction::Mov => {
                        let value = stack.pop().unwrap();
                        let value = value.into_value();
                        let Value::Integer(value) = value else {
                            return Err(BubbaError::VmPanic {
                                message: format!("Expected integer, found: {value:?}."),
                                location: location!(),
                            }
                            .into());
                        };

                        cx = value;

                        1
                    }
                    Instruction::Multiply => {
                        let b = stack.pop().unwrap();
                        let a = stack.pop().unwrap();

                        let c = a.into_value() * b.into_value();
                        stack.push(c.into());

                        1
                    }
                    Instruction::NewList(n) => {
                        let ty = stack.pop().unwrap();
                        let ty: ValueType = ty.into_value().try_into()?;

                        let ty = new_ref!(ValueType, ty);

                        let mut values = Vec::with_capacity(*n);

                        for _i in 0..*n {
                            let value = stack.pop().unwrap();
                            values.push(value.clone().into_pointer());
                        }

                        values.reverse();
                        let values = new_ref!(Vec<RefType<Value>>, values);

                        stack.push(Value::List { ty, inner: values }.into());

                        1
                    }
                    Instruction::NewMap => {
                        let map = StdHashMap::new();

                        // let key_ty = stack.pop().unwrap();
                        // let key_ty: ValueType = key_ty.into_value().try_into()?;
                        // let key_ty = new_ref!(ValueType, key_ty);

                        // let value_ty = stack.pop().unwrap();
                        // let value_ty: ValueType = value_ty.into_value().try_into()?;
                        // let value_ty = new_ref!(ValueType, value_ty);

                        stack.push(
                            Value::Map {
                                // key_ty,
                                // value_ty,
                                inner: new_ref!(StdHashMap<String, RefType<Value>>, map),
                            }
                            .into(),
                        );

                        1
                    }
                    Instruction::NewTupleEnum(n) => {
                        let variant = stack.pop().unwrap();
                        let variant: String = variant.into_value().try_into()?;

                        let path = stack.pop().unwrap();
                        let path: String = path.into_value().try_into()?;

                        let ty = stack.pop().unwrap();
                        let ty: ValueType = ty.into_value().try_into()?;

                        let ty = new_ref!(ValueType, ty);

                        let mut values: Vec<RefType<Value>> = Vec::with_capacity(*n);

                        for _i in 0..*n as i32 {
                            let value = stack.pop().unwrap();
                            values.push(value.clone().into_pointer());
                        }

                        if n > &0usize {
                            // 🚧 This is temporary until Tuples are sorted.
                            let user_enum = TupleEnum::new(variant, values[0].to_owned());
                            let user_enum = new_ref!(TupleEnum<Value>, user_enum);
                            stack.push(
                                Value::Enumeration(Enum::Tuple((ty, path), user_enum)).into(),
                            );
                        } else {
                            let user_enum = Enum::Unit(ty, path, variant);
                            stack.push(Value::Enumeration(user_enum).into());
                        }

                        1
                    }
                    Instruction::NewStruct(n) => {
                        let name = stack.pop().unwrap();
                        let name: String = name.into_value().try_into()?;

                        let ty = stack.pop().unwrap();
                        let ty: ValueType = ty.into_value().try_into()?;
                        let ty = new_ref!(ValueType, ty);

                        let mut instance = Struct::new(name, &ty);

                        for _i in 0..*n as i32 {
                            let name = stack.pop().unwrap();
                            let value = stack.pop().unwrap();

                            instance.define_field(
                                name.clone().into_value().to_inner_string(),
                                value.into_value(),
                            );
                        }

                        stack.push(Value::Struct(instance).into());

                        1
                    }
                    Instruction::Not => {
                        let value = stack.pop().unwrap();
                        let value: bool = value.into_value().try_into()?;

                        stack.push(Value::Boolean(!value).into());

                        1
                    }
                    Instruction::Or => {
                        let b = stack.pop().unwrap();
                        let a = stack.pop().unwrap();
                        let a: bool = a.into_value().try_into()?;
                        let b: bool = b.into_value().try_into()?;

                        let c = Value::Boolean(a || b);
                        stack.push(c.into());

                        1
                    }
                    Instruction::Out(stream) => {
                        let value = stack.pop().unwrap();
                        let value = value.into_value().to_inner_string();
                        let value = value.replace("\\n", "\n");

                        match stream {
                            0 => {
                                print!("{}", Colour::Green.paint(format!("{value}")));
                                std::io::Write::flush(&mut std::io::stdout()).unwrap();
                            }
                            1 => {
                                eprint!("{value}");
                                std::io::Write::flush(&mut std::io::stderr()).unwrap();
                            }
                            _ => {
                                return Err::<RefType<Value>, Error>(
                                    BubbaError::VmPanic {
                                        message: format!("Unknown stream: {stream}."),
                                        location: location!(),
                                    }
                                    .into(),
                                )
                            }
                        };

                        1
                    }
                    Instruction::PluginNew(arg_count) => {
                        let plugin_root = stack.pop().unwrap();
                        let plugin_root: String = plugin_root.into_value().try_into()?;

                        let mut args = Vec::with_capacity(*arg_count);
                        for _ in 0..*arg_count {
                            let arg = stack.pop().unwrap();
                            args.push(arg.into_value().into());
                        }

                        let library_path = RawLibrary::path_in_directory(
                            Path::new(&format!(
                                "{}/extensions/{}/lib",
                                self.home.display(),
                                plugin_root,
                            )),
                            plugin_root.as_str(),
                            LibrarySuffix::NoSuffix,
                        );
                        let root_module = (|| {
                            let header = lib_header_from_path(&library_path)?;
                            header.init_root_module::<PluginModRef>()
                        })()
                        .map_err(|e| {
                            if self.backtrace {
                                eprintln!("{self:?}");
                                print_stack(&stack, fp);
                                print_instrs(ip, &program, &self.instrs, &self.source_map);
                            }
                            BubbaError::VmPanic {
                                message: format!("Plug-in error: {e}."),
                                location: location!(),
                            }
                        })?;

                        let ctor = root_module.new();
                        let plugin = ctor(self.lambda_sender.clone().into(), args.into()).unwrap();
                        let name = plugin.name().to_string();
                        let plugin = new_ref!(PluginType, plugin);
                        let value = Value::Plugin((name, plugin));
                        stack.push(value.into());

                        1
                    }
                    Instruction::Pop => {
                        stack.pop();

                        1
                    }
                    Instruction::Push(value) => {
                        stack.push(value.clone().into());

                        1
                    }
                    Instruction::PushArgs => {
                        let value = self.args.clone();
                        stack.push(value.into());

                        1
                    }
                    Instruction::Return => {
                        let result = stack.pop().unwrap();

                        self.captures = None;

                        // Clear the stack up to the frame pointer.
                        while stack.len() > fp + 1 {
                            stack.pop();
                        }

                        // reset the frame pointer
                        fp = match stack.pop().unwrap().into_value() {
                            Value::Integer(fp) => fp as usize,
                            Value::Empty => {
                                // This is how we escape the infinite loop.
                                return Ok(result.into_pointer());
                            }
                            _ => {
                                return Err(BubbaError::VmPanic {
                                    message: format!(
                                        "Expected an integer, but got: {:?}.",
                                        stack.pop().unwrap()
                                    ),
                                    location: location!(),
                                }
                                .into());
                            }
                        };

                        let new_ip = stack.pop().unwrap();
                        let new_ip: isize = new_ip.into_value().try_into()?;

                        let frame_size = stack.pop().unwrap();
                        let frame_size: usize = frame_size.into_value().try_into()?;

                        let local_arity = stack.pop().unwrap();
                        arity = local_arity.into_value().try_into()?;

                        for _ in 0..frame_size {
                            stack.pop();
                        }

                        let frame_size = &stack[fp - 2];
                        let frame_size: usize = frame_size.clone().into_value().try_into()?;

                        // Fetch the local count from the stack, under the func addr.
                        let stack_local_count = &stack[fp - frame_size - 3 + 1];

                        // This feels pretty gross. It's trying to get the local count
                        // from the stack. If the type conversion fails, then it looks
                        // to see if there is a FubarPointer at the top of the frame
                        // and pulls the value from there.
                        //
                        // Why would there be a pointer named Fubar at the top of
                        // the frame? Well, that's what is put there when we are
                        // invoking a function on a plugin.
                        local_count = match stack_local_count.clone().into_value().try_into() {
                            Ok(local_count) => local_count,
                            Err(e) => {
                                let pointer = &stack[fp - frame_size - 3];
                                let pointer = pointer.clone().into_value();
                                if let Value::LambdaPointer {
                                    name: _,
                                    frame_size,
                                    captures: _,
                                } = pointer
                                {
                                    frame_size
                                } else {
                                    return Err(e.into());
                                }
                            }
                        };

                        stack.push(result);

                        new_ip - ip + 1
                    }
                    // The fp is pointing someplace near the end of the vec.
                    // Nominally at one past the Thonk name at the bottom of the stack.
                    // Any locals will cause the fp to be moved up, with the
                    // locals existing between the Thonk name and the fp.
                    Instruction::StoreLocal(index) => {
                        let value = stack.pop().unwrap();

                        // stack[fp - arity - local_count - 3 + index] = value;

                        // We gotta index into the stack in reverse order from the index.
                        let s = &stack[fp - arity - local_count - 3 + index];

                        // We *need* this check, otherwise we deadlock in the Pointer case.
                        // In any case, why do the work if you don't need to?
                        if value != *s {
                            // If the value from the stack is a `Value` then we can just
                            // replace it with the new value.
                            // But if it's a pointer, then we need to update the value
                            // that it's pointing at.
                            match s {
                                StackValue::Value(_) => {
                                    stack[fp - arity - local_count - 3 + index] = value;
                                }
                                StackValue::Pointer(p) => {
                                    let mut s = s_write!(p);
                                    let value = value.into_value();
                                    *s = value;
                                }
                            }
                        }

                        1
                    }
                    Instruction::StringLength => {
                        let string = stack.pop().unwrap().into_value();
                        let Value::String(string) = string else {
                            return Err(BubbaError::VmPanic {
                                message: format!("Expected a string, but got: {string:?}."),
                                location: location!(),
                            }
                            .into());
                        };
                        let len = unicode_segmentation::UnicodeSegmentation::graphemes(
                            string.as_str(),
                            true,
                        )
                        .collect::<Vec<&str>>()
                        .len();
                        stack.push(Value::Integer(len as DwarfInteger).into());

                        1
                    }
                    Instruction::StringReplace => {
                        let replace = stack.pop().unwrap().into_value();
                        let needle = stack.pop().unwrap().into_value();
                        let haystack = stack.pop().unwrap().into_value();

                        let Value::String(replace) = replace else {
                            return Err(BubbaError::VmPanic {
                                message: format!("Expected a string, but got: {replace:?}."),
                                location: location!(),
                            }
                            .into());
                        };
                        let Value::String(needle) = needle else {
                            return Err(BubbaError::VmPanic {
                                message: format!("Expected a string, but got: {needle:?}."),
                                location: location!(),
                            }
                            .into());
                        };
                        let Value::String(haystack) = haystack else {
                            return Err(BubbaError::VmPanic {
                                message: format!("Expected a string, but got: {haystack:?}."),
                                location: location!(),
                            }
                            .into());
                        };

                        let result = haystack.replace(needle.as_str(), replace.as_str());

                        stack.push(Value::String(result).into());

                        1
                    }
                    Instruction::Subtract => {
                        let b = stack.pop().unwrap();
                        let a = stack.pop().unwrap();
                        let c = a.clone().into_value() - b.clone().into_value();

                        stack.push(c.into());

                        1
                    }
                    Instruction::TestEqual => {
                        let b = stack.pop().unwrap();
                        let a = stack.pop().unwrap();
                        let a = a.into_value();
                        let b = b.into_value();
                        stack.push(Value::Boolean(a == b).into());

                        1
                    }
                    Instruction::TestGreaterThan => {
                        let b = stack.pop().unwrap();
                        let a = stack.pop().unwrap();
                        stack.push(Value::Boolean(a.into_value().gt(&b.into_value())).into());

                        1
                    }
                    Instruction::TestGreaterThanOrEqual => {
                        let b = stack.pop().unwrap();
                        let a = stack.pop().unwrap();
                        stack.push(Value::Boolean(a.into_value().gte(&b.into_value())).into());

                        1
                    }
                    Instruction::TestLessThan => {
                        let b = stack.pop().unwrap();
                        let a = stack.pop().unwrap();
                        stack.push(Value::Boolean(a.into_value().lt(&b.into_value())).into());

                        1
                    }
                    Instruction::TestLessThanOrEqual => {
                        let b = stack.pop().unwrap();
                        let a = stack.pop().unwrap();
                        stack.push(Value::Boolean(a.into_value().lte(&b.into_value())).into());

                        1
                    }
                    Instruction::TestNotEqual => {
                        let b = stack.pop().unwrap();
                        let a = stack.pop().unwrap();
                        let a = a.into_value();
                        let b = b.into_value();
                        stack.push(Value::Boolean(a != b).into());

                        1
                    }
                    Instruction::ToString => {
                        let value = stack.pop().unwrap();
                        let value = value.into_value().to_inner_string();
                        stack.push(Value::String(value).into());

                        1
                    }
                    Instruction::TypeCast(as_ty) => {
                        let Value::ValueType(as_ty) = &*s_read!(as_ty) else {
                            return Err(BubbaError::VmPanic {
                                message: format!(
                                    "Expected a ValueType, but got: {as_ty:?}.",
                                    as_ty = *as_ty
                                ),
                                location: location!(),
                            }
                            .into());
                        };

                        let lhs = stack.pop().unwrap();
                        let lhs = lhs.into_pointer();
                        let lhs = s_read!(lhs);

                        let value = match &as_ty.subtype {
                            ValueTypeEnum::List(_) => {
                                let value: Vec<RefType<Value>> = (&*lhs).try_into()?;
                                StackValue::Value(value.into())
                            }
                            ValueTypeEnum::Ty(ref ty) => {
                                let ty = self.sarzak.exhume_ty(ty).unwrap();
                                let x = match &*ty.read().unwrap() {
                                    Ty::Boolean(_) => {
                                        let value: bool = (&*lhs).try_into()?;
                                        StackValue::Value(value.into())
                                    }
                                    Ty::Float(_) => {
                                        let value: f64 = (&*lhs).try_into()?;
                                        StackValue::Value(value.into())
                                    }
                                    Ty::Integer(_) => {
                                        let value: i64 = (&*lhs).try_into()?;
                                        StackValue::Value(value.into())
                                    }
                                    Ty::ZString(_) => {
                                        let value: String = (&*lhs).try_into()?;
                                        StackValue::Value(value.into())
                                    }
                                    Ty::ZUuid(_) => {
                                        let value: uuid::Uuid = (&*lhs).try_into()?;
                                        StackValue::Value(value.into())
                                    }
                                    ref alpha => {
                                        return Err(BubbaError::VmPanic {
                                            message: format!("Unexpected type: {alpha:?}.",),
                                            location: location!(),
                                        }
                                        .into())
                                    }
                                };
                                x
                            }
                            ty => {
                                return Err(BubbaError::VmPanic {
                                    message: format!("Unexpected type: {ty:?}.",),
                                    location: location!(),
                                }
                                .into())
                            }
                        };

                        stack.push(value);

                        1
                    }
                    Instruction::Vom => {
                        stack.push(Value::Integer(cx as DwarfInteger).into());

                        1
                    }
                    invalid => {
                        panic!("Instruction not implemented: {invalid:?}.");
                    }
                }
            };

            ip += ip_offset;
        }
    }

    #[cfg(feature = "async")]
    fn start_task(
        &mut self,
        mut stack: &mut Vec<StackValue>,
        func_arity: usize,
        arity: usize,
        program: &Program,
    ) -> Result<()> {
        use puteketeke::AsyncTask;

        use crate::VmValueResult;

        let callee = &stack[stack.len() - func_arity - 2].clone();
        let stack_local_count = &stack[stack.len() - func_arity - 1].clone();
        let (name, callee, frame_size, local_card, stack_count): (
            String,
            isize,
            Value,
            usize,
            usize,
        ) = match stack_local_count.clone().into_value() {
            Value::Integer(arity) => {
                let callee: isize = callee.clone().into_value().try_into()?;
                let local_count = arity as usize;
                (
                    format!("{callee}({arity})"),
                    callee,
                    <usize as Into<Value>>::into(func_arity + local_count + 2),
                    local_count,
                    2,
                )
            }
            Value::LambdaPointer {
                name,
                frame_size,
                captures,
            } => {
                self.captures = Some(captures);
                let addr = self.func_map.get(&name).unwrap().0;
                let local_count = frame_size;
                (
                    name,
                    addr as isize,
                    // This is plus one because the call frame does not include
                    // the frame size.
                    <usize as Into<Value>>::into(func_arity + local_count + 1),
                    local_count,
                    1,
                )
            }
            _ => {
                return Err(BubbaError::VmPanic {
                    message: format!("Unexpected value: {stack_local_count:?}.",),
                    location: location!(),
                }
                .into())
            }
        };

        let mut new_stack = Vec::new();
        for _ in 0..func_arity + stack_count {
            new_stack.push(stack.pop().unwrap());
        }
        new_stack.reverse();

        let old_stack = &mut stack;
        let mut stack = new_stack;

        // The call stack has been setup, but we need to make room
        // for locals.
        (0..local_card).for_each(|_| {
            stack.push(Value::Empty.into());
        });

        // Push the arity
        stack.push(<usize as Into<Value>>::into(arity).into());

        // Push the call frame size so that we can clean in out quickly
        stack.push(frame_size.into());

        // Push the sentinel IP
        stack.push(Value::Empty.into());

        let fp = stack.len();
        // Push the sentinel frame pointer
        stack.push(Value::Empty.into());

        let mut vm = self.clone();

        let program = program.clone();
        let inner_name = name.clone();
        let future = async move {
            vm.inner_run(
                &inner_name,
                callee,
                fp,
                stack,
                func_arity,
                local_card,
                program,
            )
        };

        let executor = match unsafe { EXECUTOR.get() } {
            Some(executor) => executor,
            None => {
                let executor = Executor::new(self.thread_count);
                unsafe {
                    EXECUTOR.set(executor).unwrap();
                    EXECUTOR.get().unwrap()
                }
            }
        };
        let worker = executor.new_worker();
        let child_task = worker.spawn_task(future).unwrap();
        executor.start_task(&child_task);

        tracing::trace!(target: "vm", "Task started: {name} {child_task:?}.");

        let value = new_ref!(
            Value,
            Value::Task {
                name,
                running: true,
                task: new_ref!(Option<AsyncTask<'static, VmValueResult>>, Some(child_task))
            }
        );

        old_stack.push(value.into());

        Ok(())
    }
}

// I think that this is here for the benefit of the Result type.
impl From<(FfiValue, &Value)> for Value {
    fn from((ffi_value, ty): (FfiValue, &Value)) -> Self {
        match ffi_value {
            FfiValue::Boolean(bool_) => Self::Boolean(bool_),
            FfiValue::Empty => Self::Empty,
            // FfiValue::Error(e) => Self::Error(e.into()),
            FfiValue::Float(num) => Self::Float(num),
            FfiValue::Integer(num) => Self::Integer(num),
            FfiValue::List(list) => {
                let Value::ValueType(ty) = ty else {
                    unreachable!()
                };
                let ty = ty.clone();
                let vec: Vec<_> = list
                    .into_iter()
                    .map(|v| new_ref!(Value, v.into()))
                    .collect();
                let list = std::sync::Arc::new(std::sync::RwLock::new(vec));
                Self::List {
                    ty: new_ref!(ValueType, ty),
                    inner: list,
                }
            }
            FfiValue::Map(map) => {
                let map: StdHashMap<String, _> = map
                    .0
                    .into_iter()
                    .map(|Tuple2(k, v)| (k.into(), new_ref!(Value, v.into())))
                    .collect();
                let map = std::sync::Arc::new(std::sync::RwLock::new(map));
                Self::Map { inner: map }
            }
            FfiValue::Option(option) => match option {
                ROption::RNone => Self::Empty,
                ROption::RSome(value) => {
                    <(FfiValue, &Value) as Into<Value>>::into((RBox::into_inner(value), ty))
                }
            },
            // FfiValue::ProxyType(plugin) => Self::ProxyType {
            //     module: plugin.module.into(),
            //     obj_ty: plugin.ty.into(),
            //     id: plugin.id.into(),
            //     plugin: new_ref!(PluginType, plugin.plugin),
            // },
            FfiValue::Range(range) => Self::Range(range.start..range.end),
            FfiValue::Result(result) => {
                let tuple = match result {
                    RResult::RErr(err) => TupleEnum {
                        variant: "Err".to_owned(),
                        value: new_ref!(
                            Value,
                            <(FfiValue, &Value) as Into<Value>>::into((RBox::into_inner(err), ty))
                        ),
                    },
                    RResult::ROk(ok) => TupleEnum {
                        variant: "Ok".to_owned(),
                        value: new_ref!(
                            Value,
                            <(FfiValue, &Value) as Into<Value>>::into((RBox::into_inner(ok), ty))
                        ),
                    },
                };

                let Value::ValueType(ty) = ty else {
                    unreachable!()
                };

                Value::Enumeration(Enum::Tuple(
                    (new_ref!(ValueType, ty.to_owned()), "Result".to_owned()),
                    new_ref!(TupleEnum<Value>, tuple),
                ))
            }
            FfiValue::String(str_) => Self::String(str_.into()),
            FfiValue::Struct(struct_) => Self::Struct(struct_.into()),
            FfiValue::Uuid(uuid) => Self::Uuid(uuid.into()),
            _ => panic!("Unexpected FfiValue: {ffi_value:?}."),
        }
    }
}

fn print_stack(stack: &[StackValue], fp: usize) {
    for (i, entry) in stack.iter().enumerate() {
        if i == fp {
            eprint!("\t{} ->\t", Colour::Green.bold().paint("fp"));
        } else {
            eprint!("\t     \t");
        }
        eprintln!("stack {i}:\t{}", entry);
    }
}

fn print_instrs(ip: isize, program: &Program, instrs: &[Instruction], source_map: &[Span]) {
    for iip in 0.max(ip - 3)..(instrs.len() as isize).min(ip + 3isize) {
        let instr = &instrs[iip as usize];

        let src = if let Some(source) = program.get_source() {
            let span = source_map[iip as usize].clone();
            if span.end <= source.len() {
                &source[span]
            } else {
                ""
            }
        } else {
            ""
        };

        if ip == iip {
            eprintln!(
                "<{:08x}:\t{instr}\t\t<- {}\t{}",
                iip,
                Colour::Purple.bold().paint("ip"),
                Colour::White.dimmed().paint(src)
            );
        } else {
            eprintln!(
                "<{:08x}:\t{instr}\t\t\t{}",
                iip,
                Colour::White.dimmed().paint(src)
            );
        }
    }
}

#[cfg(test)]
mod tests {
    use std::env;
    #[cfg(feature = "tracy")]
    use tracy_client::Client;

    use crate::{
        bubba::instr::Thonk,
        dwarf::{DwarfFloat, DwarfInteger},
        interpreter::{initialize_interpreter, PrintableValueType},
        lu_dog::ObjectStore as LuDogStore,
        Context,
    };

    use super::*;

    const VERSION: &str = env!("CARGO_PKG_VERSION");
    pub const BUILD_TIME: &str = include!(concat!(env!("OUT_DIR"), "/timestamp.txt"));

    #[test]
    fn instr_constant() {
        let mut thonk = Thonk::new("test".to_string());

        thonk.add_instruction(Instruction::Push(42.into()), None);
        println!("{}", thonk);

        let mut program = Program::new(VERSION.to_owned(), BUILD_TIME.to_owned());
        program.add_thonk(thonk);
        program.add_symbol(
            "STRING".to_owned(),
            Value::ValueType(
                (*s_read!(ValueType::new_empty(true, &mut LuDogStore::new()))).clone(),
            ),
        );

        #[cfg(feature = "async")]
        let mut vm = VM::new(&program, &[], &PathBuf::new(), 1, true);
        #[cfg(not(feature = "async"))]
        let mut vm = VM::new(&program, &[], &PathBuf::new());
        let result = vm.invoke("test", &[]);
        println!("{:?}", result);
        println!("{:?}", vm);

        assert!(result.is_err());
    }

    #[test]
    fn test_instr_return() {
        let mut thonk = Thonk::new("test".to_string());

        thonk.add_instruction(Instruction::Push(42.into()), None);
        thonk.add_instruction(Instruction::Return, None);
        println!("{}", thonk);

        let mut program = Program::new(VERSION.to_owned(), BUILD_TIME.to_owned());
        program.add_thonk(thonk);
        program.add_symbol(
            "STRING".to_owned(),
            Value::ValueType(
                (*s_read!(ValueType::new_empty(true, &mut LuDogStore::new()))).clone(),
            ),
        );

        #[cfg(feature = "async")]
        let mut vm = VM::new(&program, &[], &PathBuf::new(), 1, true);
        #[cfg(not(feature = "async"))]
        let mut vm = VM::new(&program, &[], &PathBuf::new());
        let result = vm.invoke("test", &[]);
        println!("{:?}", result);
        println!("{:?}", vm);

        // assert!(vm.stack.is_empty());

        assert!(result.is_ok());

        let as_int: DwarfInteger = (&*s_read!(result.unwrap())).try_into().unwrap();
        assert_eq!(as_int, 42);

        // let mut frame = vm.frames.pop();
        // assert_eq!(frame.ip, 2);
    }

    #[test]
    fn test_instr_add() {
        let mut thonk = Thonk::new("test".to_string());

        thonk.add_instruction(Instruction::Push(69.into()), None);
        thonk.add_instruction(Instruction::Push(42.into()), None);
        thonk.add_instruction(Instruction::Add, None);
        thonk.add_instruction(Instruction::Return, None);
        println!("{}", thonk);

        let mut program = Program::new(VERSION.to_owned(), BUILD_TIME.to_owned());
        program.add_thonk(thonk);
        program.add_symbol(
            "STRING".to_owned(),
            Value::ValueType(
                (*s_read!(ValueType::new_empty(true, &mut LuDogStore::new()))).clone(),
            ),
        );

        #[cfg(feature = "async")]
        let mut vm = VM::new(&program, &[], &PathBuf::new(), 1, true);
        #[cfg(not(feature = "async"))]
        let mut vm = VM::new(&program, &[], &PathBuf::new());
        let result = vm.invoke("test", &[]);
        println!("{:?}", result);
        println!("{:?}", vm);

        // assert!(vm.stack.is_empty());

        assert!(result.is_ok());

        let as_int: DwarfInteger = (&*s_read!(result.unwrap())).try_into().unwrap();
        assert_eq!(as_int, 111);

        // let mut frame = vm.frames.pop();
        // assert_eq!(frame.ip, 4);
    }

    #[test]
    fn test_instr_subtract() {
        let mut thonk = Thonk::new("test".to_string());

        thonk.add_instruction(Instruction::Push(111.into()), None);
        thonk.add_instruction(Instruction::Push(69.into()), None);
        thonk.add_instruction(Instruction::Subtract, None);
        thonk.add_instruction(Instruction::Return, None);
        println!("{}", thonk);

        let mut program = Program::new(VERSION.to_owned(), BUILD_TIME.to_owned());
        program.add_thonk(thonk);
        program.add_symbol(
            "STRING".to_owned(),
            Value::ValueType(
                (*s_read!(ValueType::new_empty(true, &mut LuDogStore::new()))).clone(),
            ),
        );

        #[cfg(feature = "async")]
        let mut vm = VM::new(&program, &[], &PathBuf::new(), 1, true);
        #[cfg(not(feature = "async"))]
        let mut vm = VM::new(&program, &[], &PathBuf::new());
        let result = vm.invoke("test", &[]);
        println!("{:?}", result);
        println!("{:?}", vm);

        // assert!(vm.stack.is_empty());

        assert!(result.is_ok());

        let as_int: DwarfInteger = (&*s_read!(result.unwrap())).try_into().unwrap();
        assert_eq!(as_int, 42);

        // assert_eq!(frame.ip, 4);
    }

    #[test]
    fn test_instr_multiply() {
        let mut thonk = Thonk::new("test".to_string());

        thonk.add_instruction(Instruction::Push(42.into()), None);
        thonk.add_instruction(Instruction::Push(69.into()), None);
        thonk.add_instruction(Instruction::Multiply, None);
        thonk.add_instruction(Instruction::Return, None);
        println!("{}", thonk);

        let mut program = Program::new(VERSION.to_owned(), BUILD_TIME.to_owned());
        program.add_thonk(thonk);
        program.add_symbol(
            "STRING".to_owned(),
            Value::ValueType(
                (*s_read!(ValueType::new_empty(true, &mut LuDogStore::new()))).clone(),
            ),
        );

        #[cfg(feature = "async")]
        let mut vm = VM::new(&program, &[], &PathBuf::new(), 1, true);
        #[cfg(not(feature = "async"))]
        let mut vm = VM::new(&program, &[], &PathBuf::new());
        let result = vm.invoke("test", &[]);
        println!("{:?}", result);
        println!("{:?}", vm);

        // assert!(vm.stack.is_empty());

        assert!(result.is_ok());

        let as_int: DwarfInteger = (&*s_read!(result.unwrap())).try_into().unwrap();
        assert_eq!(as_int, 2898);
    }

    #[test]
    fn test_instr_less_than_or_equal() {
        // False Case
        let mut thonk = Thonk::new("test".to_string());

        thonk.add_instruction(Instruction::Push(111.into()), None);
        thonk.add_instruction(Instruction::Push(69.into()), None);
        thonk.add_instruction(Instruction::TestLessThanOrEqual, None);
        thonk.add_instruction(Instruction::Return, None);
        println!("{}", thonk);

        let mut program = Program::new(VERSION.to_owned(), BUILD_TIME.to_owned());
        program.add_thonk(thonk);
        program.add_symbol(
            "STRING".to_owned(),
            Value::ValueType(
                (*s_read!(ValueType::new_empty(true, &mut LuDogStore::new()))).clone(),
            ),
        );

        #[cfg(feature = "async")]
        let mut vm = VM::new(&program, &[], &PathBuf::new(), 1, true);
        #[cfg(not(feature = "async"))]
        let mut vm = VM::new(&program, &[], &PathBuf::new());
        let result = vm.invoke("test", &[]);
        println!("{:?}", result);
        println!("{:?}", vm);

        // assert!(vm.stack.is_empty());

        assert!(result.is_ok());

        let as_bool: bool = (&*s_read!(result.unwrap())).try_into().unwrap();
        assert!(!as_bool);

        // assert_eq!(frame.ip, 4);

        // True case: less than
        let mut thonk = Thonk::new("test".to_string());

        thonk.add_instruction(Instruction::Push(42.into()), None);
        thonk.add_instruction(Instruction::Push(69.into()), None);
        thonk.add_instruction(Instruction::TestLessThanOrEqual, None);
        thonk.add_instruction(Instruction::Return, None);
        println!("{}", thonk);

        let mut program = Program::new(VERSION.to_owned(), BUILD_TIME.to_owned());
        program.add_thonk(thonk);
        program.add_symbol(
            "STRING".to_owned(),
            Value::ValueType(
                (*s_read!(ValueType::new_empty(true, &mut LuDogStore::new()))).clone(),
            ),
        );

        #[cfg(feature = "async")]
        let mut vm = VM::new(&program, &[], &PathBuf::new(), 1, true);
        #[cfg(not(feature = "async"))]
        let mut vm = VM::new(&program, &[], &PathBuf::new());
        let result = vm.invoke("test", &[]);
        println!("{:?}", result);
        println!("{:?}", vm);

        // assert!(vm.stack.is_empty());

        assert!(result.is_ok());

        let as_bool: bool = (&*s_read!(result.unwrap())).try_into().unwrap();
        assert!(as_bool);

        // assert_eq!(frame.ip, 4);

        // True case: equal
        let mut thonk = Thonk::new("test".to_string());

        thonk.add_instruction(Instruction::Push(42.into()), None);
        thonk.add_instruction(Instruction::Push(42.into()), None);
        thonk.add_instruction(Instruction::TestLessThanOrEqual, None);
        thonk.add_instruction(Instruction::Return, None);
        println!("{}", thonk);

        let mut program = Program::new(VERSION.to_owned(), BUILD_TIME.to_owned());
        program.add_thonk(thonk);
        program.add_symbol(
            "STRING".to_owned(),
            Value::ValueType(
                (*s_read!(ValueType::new_empty(true, &mut LuDogStore::new()))).clone(),
            ),
        );

        #[cfg(feature = "async")]
        let mut vm = VM::new(&program, &[], &PathBuf::new(), 1, true);
        #[cfg(not(feature = "async"))]
        let mut vm = VM::new(&program, &[], &PathBuf::new());
        let result = vm.invoke("test", &[]);
        println!("{:?}", result);
        println!("{:?}", vm);

        // assert!(vm.stack.is_empty());

        assert!(result.is_ok());

        let as_bool: bool = (&*s_read!(result.unwrap())).try_into().unwrap();
        assert!(as_bool);

        // let mut frame = vm.frames.pop();
        // assert_eq!(frame.ip, 4);
    }

    #[test]
    fn test_instr_jump_if_false() {
        let mut thonk = Thonk::new("test".to_string());

        thonk.add_instruction(Instruction::Push(69.into()), None);
        thonk.add_instruction(Instruction::Push(42.into()), None);
        thonk.add_instruction(Instruction::TestLessThanOrEqual, None);
        thonk.add_instruction(Instruction::JumpIfFalse(2), None);
        thonk.add_instruction(
            Instruction::Push(Value::String("epic fail!".to_string())),
            None,
        );
        thonk.add_instruction(Instruction::Return, None);
        thonk.add_instruction(
            Instruction::Push(Value::String("you rock!".to_string())),
            None,
        );
        thonk.add_instruction(Instruction::Return, None);
        println!("{}", thonk);

        let mut program = Program::new(VERSION.to_owned(), BUILD_TIME.to_owned());
        program.add_thonk(thonk);
        program.add_symbol(
            "STRING".to_owned(),
            Value::ValueType(
                (*s_read!(ValueType::new_empty(true, &mut LuDogStore::new()))).clone(),
            ),
        );

        #[cfg(feature = "async")]
        let mut vm = VM::new(&program, &[], &PathBuf::new(), 1, true);
        #[cfg(not(feature = "async"))]
        let mut vm = VM::new(&program, &[], &PathBuf::new());
        let result = vm.invoke("test", &[]);
        println!("{:?}", result);
        println!("{:?}", vm);

        // assert!(vm.stack.is_empty());

        assert!(result.is_ok());

        let result: String = (&*s_read!(result.unwrap())).try_into().unwrap();
        assert_eq!(result, "you rock!");

        // let mut frame = vm.frames.pop();
        // assert_eq!(frame.ip, 8);
    }

    #[test]
    fn test_instr_store_fetch_local() {
        // Simple
        let mut thonk = Thonk::new("test".to_string());
        thonk.increment_frame_size();
        thonk.add_instruction(Instruction::Push(42.into()), None);
        thonk.add_instruction(Instruction::StoreLocal(0), None);
        thonk.add_instruction(Instruction::FetchLocal(0), None);
        thonk.add_instruction(Instruction::Return, None);
        println!("{}", thonk);
        let mut program = Program::new(VERSION.to_owned(), BUILD_TIME.to_owned());
        program.add_thonk(thonk);
        program.add_symbol(
            "STRING".to_owned(),
            Value::ValueType(
                (*s_read!(ValueType::new_empty(true, &mut LuDogStore::new()))).clone(),
            ),
        );

        #[cfg(feature = "async")]
        let mut vm = VM::new(&program, &[], &PathBuf::new(), 1, true);
        #[cfg(not(feature = "async"))]
        let mut vm = VM::new(&program, &[], &PathBuf::new());

        let result = vm.invoke("test", &[]);
        println!("{:?}", result);
        println!("{:?}", vm);

        // assert!(vm.stack.is_empty());

        assert!(result.is_ok());

        let result: DwarfInteger = (&*s_read!(result.unwrap())).try_into().unwrap();
        assert_eq!(result, 42);
    }

    #[test]
    fn test_instr_field() {
        use crate::{
            chacha::value::Struct,
            lu_dog::{Field, ValueType, WoogStruct},
            PATH_ROOT,
        };
        use sarzak::sarzak::{ObjectStore as SarzakStore, Ty, MODEL as SARZAK_MODEL};

        #[cfg(feature = "tracy")]
        Client::start();

        let sarzak = SarzakStore::from_bincode(SARZAK_MODEL).unwrap();

        let ctx = Context::default();
        let struct_ty = {
            let mut lu_dog = s_write!(ctx.lu_dog);

            // We need to create a WoogStruct and add some fields to it
            let foo = WoogStruct::new(
                "Foo".to_owned(),
                PATH_ROOT.to_owned(),
                None,
                None,
                &mut lu_dog,
            );
            // let _ = WoogItem::new_woog_struct(source, &mt, lu_dog);
            let struct_ty = ValueType::new_woog_struct(true, &foo, &mut lu_dog);
            let ty = Ty::new_integer(&sarzak);
            let ty = ValueType::new_ty(true, &ty, &mut lu_dog);
            let _ = Field::new("bar".to_owned(), &foo, &ty, &mut lu_dog);
            let ty = Ty::new_float(&sarzak);
            let ty = ValueType::new_ty(true, &ty, &mut lu_dog);
            let _ = Field::new("baz".to_owned(), &foo, &ty, &mut lu_dog);
            struct_ty
        };

        let ty = Ty::new_z_string(&sarzak);
        let ty = ValueType::new_ty(true, &ty, &mut s_write!(ctx.lu_dog));
        let ty = Value::ValueType((*s_read!(ty)).clone());

        // Now we need an instance.
        let dwarf_home = env::var("DWARF_HOME")
            .unwrap_or_else(|_| {
                let mut home = env::var("HOME").unwrap();
                home.push_str("/.dwarf");
                home
            })
            .into();

        let ctx = initialize_interpreter(2, dwarf_home, ctx).unwrap();
        let ty_name = PrintableValueType(false, struct_ty.clone(), ctx.models());
        let mut foo_inst = Struct::new(ty_name.to_string(), &struct_ty);
        foo_inst.define_field("bar", 42.into());
        foo_inst.define_field("baz", std::f64::consts::PI.into());

        let mut thonk = Thonk::new("test".to_string());
        thonk.add_instruction(Instruction::Push(Value::Struct(foo_inst)), None);
        thonk.add_instruction(Instruction::Push("baz".into()), None);
        thonk.add_instruction(Instruction::FieldRead, None);
        thonk.add_instruction(Instruction::Return, None);
        println!("{}", thonk);
        let mut program = Program::new(VERSION.to_owned(), BUILD_TIME.to_owned());
        program.add_thonk(thonk);

        program.add_symbol("STRING".to_owned(), ty);

        #[cfg(feature = "async")]
        let mut vm = VM::new(&program, &[], &PathBuf::new(), 1, true);
        #[cfg(not(feature = "async"))]
        let mut vm = VM::new(&program, &[], &PathBuf::new());

        let result = vm.invoke("test", &[]);
        println!("{:?}", result);
        println!("{:?}", vm);

        // assert!(vm.stack.is_empty());

        assert!(result.is_ok());

        let result: DwarfFloat = (&*s_read!(result.unwrap())).try_into().unwrap();
        assert_eq!(result, std::f64::consts::PI);
    }
}
