use std::{env, path::PathBuf};

use ansi_term::Colour;
use lazy_static::lazy_static;
use parking_lot::Mutex;
use test_log::test;

use dwarf::{
    bubba::{
        compiler::{compile, BubbaCompilerErrorReporter},
        error::BubbaErrorReporter,
        value::Value as BubbaValue,
        VM,
    },
    dwarf::{new_lu_dog, parse_dwarf},
    s_read,
    sarzak::{ObjectStore as SarzakStore, MODEL as SARZAK_MODEL},
};

const NUM_THREADS: usize = 4;

lazy_static! {
    static ref EXEC_MUTEX: Mutex<()> = Mutex::new(());
}

#[cfg(feature = "print-std-out")]
compile_error!("The tests don't function with the print-std-out feature enabled.");

fn output_diffs(expected: &str, found: &str, test: &str) -> Result<(), ()> {
    let mut diff_count = 0;
    let mut diff = String::new();
    for line in diff::lines(expected, found) {
        match line {
            diff::Result::Left(expected) => {
                if expected.starts_with("[31mError:[0m Unexpected token in input, expected") {
                    continue;
                }
                diff_count += 1;
                diff += &format!("{} {expected}\n", Colour::Green.paint("+++"));
            }
            diff::Result::Right(found) => {
                if found.starts_with("[31mError:[0m Unexpected token in input, expected") {
                    continue;
                }
                diff += &format!("{} {found}\n", Colour::Red.paint("---"));
            }
            diff::Result::Both(a, _) => eprintln!("    {a}"),
        }
    }

    if diff_count > 0 {
        eprintln!(
            "{}",
            Colour::Red.paint(format!(
                "stderr does not match .stderr file for test {test}"
            ))
        );
        eprintln!("Expected:\n{expected}");
        eprintln!("Found:\n{found}");
        eprintln!("Diff:\n{diff}");
        Err(())
    } else {
        Ok(())
    }
}

fn diff_with_file(path: &str, test: &str, found: &str) -> Result<(), ()> {
    let path = PathBuf::from(path);
    let stdout = std::fs::read_to_string(path).unwrap().trim().to_owned();
    if stdout == found {
        // The output matches the .stdout file. We pass the test.
        Ok(())
    } else {
        // The output does not match the .stdout file -- do a diff.
        output_diffs(&stdout, found, test)
    }
}

fn run_program(test: &str, program: &str, cwd: &PathBuf) -> Result<(BubbaValue, String), String> {
    let _guard = EXEC_MUTEX.lock();
    let sarzak = SarzakStore::from_bincode(SARZAK_MODEL).unwrap();
    let dwarf_home = env::var("DWARF_HOME")
        .unwrap_or_else(|_| {
            let mut home = env::var("HOME").unwrap();
            home.push_str("/.dwarf");
            home
        })
        .into();

    let ast = match parse_dwarf(test, program) {
        Ok(ast) => ast,
        Err(e) => match *e {
            dwarf::dwarf::error::DwarfError::Parse { error, ast: _ } => {
                let error = error.trim();
                eprintln!("{error}");
                return Err(error.to_owned());
            }
            e => {
                eprintln!("{e:?}");
                return Err(e.to_string());
            }
        },
    };

    let ctx = match new_lu_dog(
        test.to_owned(),
        Some((program.to_owned(), &ast)),
        &dwarf_home,
        cwd,
        &sarzak,
    ) {
        Ok(lu_dog) => lu_dog,
        Err(e) => {
            eprintln!(
                "{}",
                e.iter()
                    .map(|e| {
                        format!(
                            "{}",
                            // This one is uber.
                            dwarf::dwarf::error::DwarfErrorReporter(e, true)
                        )
                    })
                    .collect::<Vec<_>>()
                    .join("\n")
                    .trim()
            );

            let errors = e
                .iter()
                .map(|e| {
                    format!(
                        "{}",
                        // This one is not uber.
                        dwarf::dwarf::error::DwarfErrorReporter(e, false)
                    )
                })
                .collect::<Vec<_>>()
                .join("\n")
                .trim()
                .to_owned();

            return Err(errors);
        }
    };

    let binary = match compile(&ctx) {
        Ok(program) => program,
        Err(e) => {
            let error = format!(
                "Unable to compile program:\n{}",
                BubbaCompilerErrorReporter(&e, true, program, test)
            )
            .trim()
            .to_owned();

            eprintln!("{error}");

            let error = format!(
                "Unable to compile program:\n{}",
                BubbaCompilerErrorReporter(&e, false, program, test)
            )
            .trim()
            .to_owned();

            return Err(error);
        }
    };

    #[cfg(feature = "async")]
    let mut vm = VM::new(&binary, &[], &dwarf_home, NUM_THREADS, false);
    #[cfg(not(feature = "async"))]
    let mut vm = VM::new(&program, &[], &dwarf_home, false);

    let result = match vm.invoke("main", &[]) {
        Ok(value) => {
            let value = s_read!(value).clone();

            match value {
                BubbaValue::Error(msg) => {
                    let msg = *msg;
                    let error = format!(
                        "Vm exited with:\n{}",
                        BubbaErrorReporter(&msg.into(), true, program, test)
                    )
                    .trim()
                    .to_owned();

                    eprintln!("{error}");

                    Err(error)

                    // eprintln!("{msg}");
                    // Err(msg.to_string())
                }
                _ => Ok((value, "Oops".to_owned())),
            }
        }
        Err(e) => {
            let error = format!(
                "VM exited with:\n{}",
                BubbaErrorReporter(&e, true, program, test)
            )
            .trim()
            .to_owned();

            eprintln!("{error}");

            let error = format!(
                "VM exited with:\n{}",
                BubbaErrorReporter(&e, false, program, test)
            )
            .trim()
            .to_owned();

            Err(error)
        }
    };

    result
}

// This loads the generated tests.
include!(concat!(env!("OUT_DIR"), "/vm_tests.rs"));
